@@ -54,6 +54,9 @@ static const uint32_t mlx4_ib_opcode[] = {
 	[IBV_WR_RDMA_READ]		= MLX4_OPCODE_RDMA_READ,
 	[IBV_WR_ATOMIC_CMP_AND_SWP]	= MLX4_OPCODE_ATOMIC_CS,
 	[IBV_WR_ATOMIC_FETCH_AND_ADD]	= MLX4_OPCODE_ATOMIC_FA,
+	[IBV_WR_LOCAL_INV]		= MLX4_OPCODE_LOCAL_INVAL,
+	[IBV_WR_BIND_MW]		= MLX4_OPCODE_BIND_MW,
+	[IBV_WR_SEND_WITH_INV]		= MLX4_OPCODE_SEND_INVAL,
 };
 
 static void *get_recv_wqe(struct mlx4_qp *qp, int n)
@@ -118,6 +121,40 @@ static int wq_overflow(struct mlx4_wq *wq, int nreq, struct mlx4_cq *cq)
 	return cur + nreq >= wq->max_post;
 }
 
+static void set_bind_seg(struct mlx4_wqe_bind_seg *bseg, struct ibv_send_wr *wr)
+{
+	int acc = wr->bind_mw.bind_info.mw_access_flags;
+	bseg->flags1 = 0;
+	if (acc & IBV_ACCESS_REMOTE_ATOMIC)
+		bseg->flags1 |= htonl(MLX4_WQE_MW_ATOMIC);
+	if (acc & IBV_ACCESS_REMOTE_WRITE)
+		bseg->flags1 |= htonl(MLX4_WQE_MW_REMOTE_WRITE);
+	if (acc & IBV_ACCESS_REMOTE_READ)
+		bseg->flags1 |= htonl(MLX4_WQE_MW_REMOTE_READ);
+
+	bseg->flags2 = 0;
+	if (((struct ibv_mw *)(wr->bind_mw.mw))->type == IBV_MW_TYPE_2)
+		bseg->flags2 |= htonl(MLX4_WQE_BIND_TYPE_2);
+	if (acc & IBV_ACCESS_ZERO_BASED)
+		bseg->flags2 |= htonl(MLX4_WQE_BIND_ZERO_BASED);
+
+	bseg->new_rkey = htonl(wr->bind_mw.rkey);
+	bseg->lkey = htonl(wr->bind_mw.bind_info.mr->lkey);
+	bseg->addr = htobe64((uint64_t) wr->bind_mw.bind_info.addr);
+	bseg->length = htobe64(wr->bind_mw.bind_info.length);
+}
+
+static inline void set_local_inv_seg(struct mlx4_wqe_local_inval_seg *iseg,
+		uint32_t rkey)
+{
+	iseg->mem_key	= htonl(rkey);
+
+	iseg->reserved1    = 0;
+	iseg->reserved2    = 0;
+	iseg->reserved3[0] = 0;
+	iseg->reserved3[1] = 0;
+}
+
 static inline void set_raddr_seg(struct mlx4_wqe_raddr_seg *rseg,
 				 uint64_t remote_addr, uint32_t rkey)
 {
@@ -170,21 +207,10 @@ static void set_data_seg(struct mlx4_wqe_data_seg *dseg, struct ibv_sge *sg)
 	 */
 	wmb();
 
-	dseg->byte_count = htonl(sg->length);
-}
-
-/*
- * Avoid using memcpy() to copy to BlueFlame page, since memcpy()
- * implementations may use move-string-buffer assembler instructions,
- * which do not guarantee order of copying.
- */
-static void mlx4_bf_copy(unsigned long *dst, unsigned long *src, unsigned bytecnt)
-{
-	while (bytecnt > 0) {
-		*dst++ = *src++;
-		*dst++ = *src++;
-		bytecnt -= 2 * sizeof (long);
-	}
+	if (likely(sg->length))
+		dseg->byte_count = htonl(sg->length);
+	else
+		dseg->byte_count = htonl(0x80000000);
 }
 
 int mlx4_post_send(struct ibv_qp *ibqp, struct ibv_send_wr *wr,
@@ -193,12 +219,12 @@ int mlx4_post_send(struct ibv_qp *ibqp, struct ibv_send_wr *wr,
 	struct mlx4_context *ctx;
 	struct mlx4_qp *qp = to_mqp(ibqp);
 	void *wqe;
-	struct mlx4_wqe_ctrl_seg *ctrl;
+	struct mlx4_wqe_ctrl_seg *uninitialized_var(ctrl);
 	int ind;
 	int nreq;
 	int inl = 0;
 	int ret = 0;
-	int size;
+	int size = 0;
 	int i;
 
 	pthread_spin_lock(&qp->sq.lock);
@@ -278,6 +304,27 @@ int mlx4_post_send(struct ibv_qp *ibqp, struct ibv_send_wr *wr,
 				size += sizeof (struct mlx4_wqe_raddr_seg) / 16;
 
 				break;
+			case IBV_WR_LOCAL_INV:
+				ctrl->srcrb_flags |=
+					htonl(MLX4_WQE_CTRL_STRONG_ORDER);
+				set_local_inv_seg(wqe, wr->imm_data);
+				wqe  += sizeof
+					(struct mlx4_wqe_local_inval_seg);
+				size += sizeof
+					(struct mlx4_wqe_local_inval_seg) / 16;
+				break;
+			case IBV_WR_BIND_MW:
+				ctrl->srcrb_flags |=
+					htonl(MLX4_WQE_CTRL_STRONG_ORDER);
+				set_bind_seg(wqe, wr);
+				wqe  += sizeof
+					(struct mlx4_wqe_bind_seg);
+				size += sizeof
+					(struct mlx4_wqe_bind_seg) / 16;
+				break;
+			case IBV_WR_SEND_WITH_INV:
+				ctrl->imm = htonl(wr->imm_data);
+				break;
 
 			default:
 				/* No extra segments required for sends */
@@ -289,12 +336,31 @@ int mlx4_post_send(struct ibv_qp *ibqp, struct ibv_send_wr *wr,
 			set_datagram_seg(wqe, wr);
 			wqe  += sizeof (struct mlx4_wqe_datagram_seg);
 			size += sizeof (struct mlx4_wqe_datagram_seg) / 16;
+
+			if (wr->send_flags & IBV_SEND_IP_CSUM) {
+				if (!(qp->qp_cap_cache & MLX4_CSUM_SUPPORT_UD_OVER_IB)) {
+					ret = EINVAL;
+					*bad_wr = wr;
+					goto out;
+				}
+				ctrl->srcrb_flags |= htonl(MLX4_WQE_CTRL_IP_HDR_CSUM |
+							   MLX4_WQE_CTRL_TCP_UDP_CSUM);
+			}
 			break;
 
 		case IBV_QPT_RAW_PACKET:
 			/* For raw eth, the MLX4_WQE_CTRL_SOLICIT flag is used
 			 * to indicate that no icrc should be calculated */
 			ctrl->srcrb_flags |= htonl(MLX4_WQE_CTRL_SOLICIT);
+			if (wr->send_flags & IBV_SEND_IP_CSUM) {
+				if (!(qp->qp_cap_cache & MLX4_CSUM_SUPPORT_RAW_OVER_ETH)) {
+					ret = EINVAL;
+					*bad_wr = wr;
+					goto out;
+				}
+				ctrl->srcrb_flags |= htonl(MLX4_WQE_CTRL_IP_HDR_CSUM |
+							   MLX4_WQE_CTRL_TCP_UDP_CSUM);
+			}
 			break;
 
 		default:
@@ -407,7 +473,8 @@ out:
 
 	if (nreq == 1 && inl && size > 1 && size <= ctx->bf_buf_size / 16) {
 		ctrl->owner_opcode |= htonl((qp->sq.head & 0xffff) << 8);
-		*(uint32_t *) ctrl->reserved |= qp->doorbell_qpn;
+
+		ctrl->bf_qpn |= qp->doorbell_qpn;
 		/*
 		 * Make sure that descriptor is written to memory
 		 * before writing to BlueFlame page.
@@ -434,7 +501,8 @@ out:
 		 */
 		wmb();
 
-		*(uint32_t *) (ctx->uar + MLX4_SEND_DOORBELL) = qp->doorbell_qpn;
+		mmio_writel((unsigned long)(ctx->uar + MLX4_SEND_DOORBELL),
+			    qp->doorbell_qpn);
 	}
 
 	if (nreq)
